#include "Processor.h"

ULONG Locked = 0;

typedef enum _KOBJECTS
{
	EventNotificationObject,
  EventSynchronizationObject,
  MutantObject,
  ProcessObject,
  QueueObject,
  SemaphoreObject,
  ThreadObject,
  GateObject,
  TimerNotificationObject,
  TimerSynchronizationObject,
  Spare2Object,
  Spare3Object,
  Spare4Object,
  Spare5Object,
  Spare6Object,
  Spare7Object,
  Spare8Object,
  Spare9Object,
  ApcObject,
  DpcObject,
  DeviceQueueObject,
  EventPairObject,
  InterruptObject,
  ProfileObject,
  ThreadedDpcObject,
  MaximumKernelObject
} KOBJECTS;

typedef struct _DPC_PARAMS
{
	KDPC           Dpc;
	PCALLBACK_PROC Proc;
	PVOID          Param;
	PKEVENT        SyncEvent;
	BOOLEAN        Syncronous;
} DPC_PARAMS, *PDPC_PARAMS;


void
Ioport_KiInitializeDpc(
			IN PRKDPC Dpc,
			IN PKDEFERRED_ROUTINE DeferredRoutine,
			IN OPTIONAL PVOID DeferredContext,
			IN KOBJECTS DpcType
			)
{
	Dpc->Type = DpcType;
  Dpc->DpcData = NULL;
	Dpc->Importance = 1;
	Dpc->Number = 0;
  Dpc->DeferredRoutine = DeferredRoutine;
  Dpc->DeferredContext = DeferredContext;
}

void
Ioport_KeInitializeDpc(
			OUT PRKDPC Dpc,
			IN PKDEFERRED_ROUTINE DeferredRoutine,
			IN OPTIONAL PVOID DeferredContext
			)
{
	Ioport_KiInitializeDpc(Dpc, DeferredRoutine, DeferredContext, DpcObject);
}

void
Ioport_KeInitializeEvent(
			OUT PRKEVENT Event,
			IN EVENT_TYPE Type,
			IN BOOLEAN State
			)
{
  Event->Header.Type = (UCHAR)Type;
  Event->Header.Absolute = 0;
  Event->Header.Size = sizeof(KEVENT) / sizeof(LONG);
  Event->Header.SignalState = State;
	InitializeListHead(&Event->Header.WaitListHead);
}


void
Ioport_KeSetTargetProcessorDpc(
    IN OUT PRKDPC Dpc,
    IN CCHAR Number
    )
{
    ASSERT_DPC(Dpc);
    Dpc->Number = 64 + Number;
}


void DpcRoutine(
    IN PKDPC Dpc,
    IN PVOID DeferredContext,
    IN PVOID SystemArgument1,
    IN PVOID SystemArgument2
    )
{
	PDPC_PARAMS Params;

	UNREFERENCED_PARAMETER(Dpc);
	UNREFERENCED_PARAMETER(SystemArgument1);
	UNREFERENCED_PARAMETER(SystemArgument2);

	Params = DeferredContext;

	Params->Proc(Params->Param);

	if (Params->Syncronous) KeSetEvent(Params->SyncEvent, IO_NO_INCREMENT, FALSE);

	ExFreePool(Params);
}

void CallToProcessor(
				  IN CCHAR          Cpu,
				  IN PCALLBACK_PROC Proc,
				  IN PVOID          Param,
				  IN BOOLEAN        Syncronous
				  )
{
	KEVENT       SyncEvent;
	PDPC_PARAMS  DpcParams;

	DpcParams = ExAllocatePool(NonPagedPool, sizeof(DPC_PARAMS));

	if (DpcParams)
	{
		Ioport_KeInitializeEvent(&SyncEvent, NotificationEvent, FALSE);

		DpcParams->Syncronous = Syncronous;
		DpcParams->Param      = Param;
		DpcParams->Proc       = Proc;
		DpcParams->SyncEvent  = &SyncEvent;

		Ioport_KeInitializeDpc(&DpcParams->Dpc, DpcRoutine, DpcParams);

		Ioport_KeSetTargetProcessorDpc(&DpcParams->Dpc, Cpu);

		KeInsertQueueDpc(&DpcParams->Dpc, NULL, NULL);

		if (Syncronous) KeWaitForSingleObject(&SyncEvent, Executive, KernelMode, FALSE, NULL);
	}	
}

void CallToAllProcessors(
				  IN PCALLBACK_PROC Proc,
				  IN PVOID          Param,
				  IN BOOLEAN        NoSelfCall,
				  IN BOOLEAN        Syncronous
				  )
{
	KIRQL        OldIrql;
	CCHAR        Num;
	ULONG        MyProc;

	if (KeNumberProcessors == 1)
	{
		if (!NoSelfCall)
		{
			OldIrql = (KIRQL)ReadCR8();
			WriteCR8(DISPATCH_LEVEL);
			
			Proc(Param);

			WriteCR8(OldIrql);
		}
	} else
	{
		MyProc = KeGetCurrentProcessorNumber();

		for (Num = 0; Num < KeNumberProcessors; Num++)
		{
			if (!NoSelfCall || (ULONG)Num != MyProc) CallToProcessor(Num, Proc, Param, Syncronous);
		}
	}		
}

extern void LockProc(ULONG l);

void LockProcWithoutParams()
{
	LockProc(Locked);
}

KIRQL GlobalLock()
{
	KIRQL Result = (KIRQL)ReadCR8();

	WriteCR8(DISPATCH_LEVEL);
	
	Locked++;
	
	if (KeNumberProcessors > 1)
		CallToAllProcessors((PCALLBACK_PROC)LockProcWithoutParams, NULL, TRUE, FALSE);

	return Result;
}

void GlobalUnlock(
				IN KIRQL OldIrql
				)
{
	Locked--;

	WriteCR8(OldIrql);
}