#include <ntifs.h>
#include "Processor.h"
#include "ProcessList.h"

PProcessList wLastItem = NULL;

// from EPROCESS struct
ULONG UniqueProcessId = 0x180;
ULONG ActiveProcessLinks = 0x188;

PDEVICE_OBJECT deviceObject;
PLIST_ENTRY PsActiveProcessHead;

//
// This structure is used by the debugger for all targets
// It is the same size as DBGKD_DATA_HEADER on all systems
//
typedef struct _DBGKD_DEBUG_DATA_HEADER64 {

	//
	// Link to other blocks
	//

	LIST_ENTRY64 List;

	//
	// This is a unique tag to identify the owner of the block.
	// If your component only uses one pool tag, use it for this, too.
	//

	ULONG           OwnerTag;

	//
	// This must be initialized to the size of the data block,
	// including this structure.
	//

	ULONG           Size;

} DBGKD_DEBUG_DATA_HEADER64, *PDBGKD_DEBUG_DATA_HEADER64;

typedef struct _KDDEBUGGER_DATA64 {

	DBGKD_DEBUG_DATA_HEADER64 Header;

	//
	// Base address of kernel image
	//

	ULONG64   KernBase;

	//
	// DbgBreakPointWithStatus is a function which takes an argument
	// and hits a breakpoint.  This field contains the address of the
	// breakpoint instruction.  When the debugger sees a breakpoint
	// at this address, it may retrieve the argument from the first
	// argument register, or on x86 the eax register.
	//

	ULONG64   BreakpointWithStatus;       // address of breakpoint

	//
	// Address of the saved context record during a bugcheck
	//
	// N.B. This is an automatic in KeBugcheckEx's frame, and
	// is only valid after a bugcheck.
	//

	ULONG64   SavedContext;

	//
	// help for walking stacks with user callbacks:
	//

	//
	// The address of the thread structure is provided in the
	// WAIT_STATE_CHANGE packet.  This is the offset from the base of
	// the thread structure to the pointer to the kernel stack frame
	// for the currently active usermode callback.
	//

	USHORT  ThCallbackStack;            // offset in thread data

	//
	// these values are offsets into that frame:
	//

	USHORT  NextCallback;               // saved pointer to next callback frame
	USHORT  FramePointer;               // saved frame pointer

	//
	// pad to a quad boundary
	//
	USHORT  PaeEnabled;

	//
	// Address of the kernel callout routine.
	//

	ULONG64   KiCallUserMode;             // kernel routine

	//
	// Address of the usermode entry point for callbacks.
	//

	ULONG64   KeUserCallbackDispatcher;   // address in ntdll


	//
	// Addresses of various kernel data structures and lists
	// that are of interest to the kernel debugger.
	//

	ULONG64   PsLoadedModuleList;
	ULONG64   PsActiveProcessHead;
	ULONG64   PspCidTable;

	ULONG64   ExpSystemResourcesList;
	ULONG64   ExpPagedPoolDescriptor;
	ULONG64   ExpNumberOfPagedPools;

	ULONG64   KeTimeIncrement;
	ULONG64   KeBugCheckCallbackListHead;
	ULONG64   KiBugcheckData;

	ULONG64   IopErrorLogListHead;

	ULONG64   ObpRootDirectoryObject;
	ULONG64   ObpTypeObjectType;

	ULONG64   MmSystemCacheStart;
	ULONG64   MmSystemCacheEnd;
	ULONG64   MmSystemCacheWs;

	ULONG64   MmPfnDatabase;
	ULONG64   MmSystemPtesStart;
	ULONG64   MmSystemPtesEnd;
	ULONG64   MmSubsectionBase;
	ULONG64   MmNumberOfPagingFiles;

	ULONG64   MmLowestPhysicalPage;
	ULONG64   MmHighestPhysicalPage;
	ULONG64   MmNumberOfPhysicalPages;

	ULONG64   MmMaximumNonPagedPoolInBytes;
	ULONG64   MmNonPagedSystemStart;
	ULONG64   MmNonPagedPoolStart;
	ULONG64   MmNonPagedPoolEnd;

	ULONG64   MmPagedPoolStart;
	ULONG64   MmPagedPoolEnd;
	ULONG64   MmPagedPoolInformation;
	ULONG64   MmPageSize;

	ULONG64   MmSizeOfPagedPoolInBytes;

	ULONG64   MmTotalCommitLimit;
	ULONG64   MmTotalCommittedPages;
	ULONG64   MmSharedCommit;
	ULONG64   MmDriverCommit;
	ULONG64   MmProcessCommit;
	ULONG64   MmPagedPoolCommit;
	ULONG64   MmExtendedCommit;

	ULONG64   MmZeroedPageListHead;
	ULONG64   MmFreePageListHead;
	ULONG64   MmStandbyPageListHead;
	ULONG64   MmModifiedPageListHead;
	ULONG64   MmModifiedNoWritePageListHead;
	ULONG64   MmAvailablePages;
	ULONG64   MmResidentAvailablePages;

	ULONG64   PoolTrackTable;
	ULONG64   NonPagedPoolDescriptor;

	ULONG64   MmHighestUserAddress;
	ULONG64   MmSystemRangeStart;
	ULONG64   MmUserProbeAddress;

	ULONG64   KdPrintCircularBuffer;
	ULONG64   KdPrintCircularBufferEnd;
	ULONG64   KdPrintWritePointer;
	ULONG64   KdPrintRolloverCount;

	ULONG64   MmLoadedUserImageList;

	// NT 5.1 Addition

	ULONG64   NtBuildLab;
	ULONG64   KiNormalSystemCall;

	// NT 5.0 hotfix addition

	ULONG64   KiProcessorBlock;
	ULONG64   MmUnloadedDrivers;
	ULONG64   MmLastUnloadedDriver;
	ULONG64   MmTriageActionTaken;
	ULONG64   MmSpecialPoolTag;
	ULONG64   KernelVerifier;
	ULONG64   MmVerifierData;
	ULONG64   MmAllocatedNonPagedPool;
	ULONG64   MmPeakCommitment;
	ULONG64   MmTotalCommitLimitMaximum;
	ULONG64   CmNtCSDVersion;

	// NT 5.1 Addition

	ULONG64   MmPhysicalMemoryBlock;
	ULONG64   MmSessionBase;
	ULONG64   MmSessionSize;
	ULONG64   MmSystemParentTablePage;

	// Server 2003 addition

	ULONG64   MmVirtualTranslationBase;

	USHORT    OffsetKThreadNextProcessor;
	USHORT    OffsetKThreadTeb;
	USHORT    OffsetKThreadKernelStack;
	USHORT    OffsetKThreadInitialStack;

	USHORT    OffsetKThreadApcProcess;
	USHORT    OffsetKThreadState;
	USHORT    OffsetKThreadBStore;
	USHORT    OffsetKThreadBStoreLimit;

	USHORT    SizeEProcess;
	USHORT    OffsetEprocessPeb;
	USHORT    OffsetEprocessParentCID;
	USHORT    OffsetEprocessDirectoryTableBase;

	USHORT    SizePrcb;
	USHORT    OffsetPrcbDpcRoutine;
	USHORT    OffsetPrcbCurrentThread;
	USHORT    OffsetPrcbMhz;

	USHORT    OffsetPrcbCpuType;
	USHORT    OffsetPrcbVendorString;
	USHORT    OffsetPrcbProcStateContext;
	USHORT    OffsetPrcbNumber;

	USHORT    SizeEThread;

	ULONG64   KdPrintCircularBufferPtr;
	ULONG64   KdPrintBufferSize;

	ULONG64   KeLoaderBlock;

	USHORT    SizePcr;
	USHORT    OffsetPcrSelfPcr;
	USHORT    OffsetPcrCurrentPrcb;
	USHORT    OffsetPcrContainedPrcb;

	USHORT    OffsetPcrInitialBStore;
	USHORT    OffsetPcrBStoreLimit;
	USHORT    OffsetPcrInitialStack;
	USHORT    OffsetPcrStackLimit;

	USHORT    OffsetPrcbPcrPage;
	USHORT    OffsetPrcbProcStateSpecialReg;
	USHORT    GdtR0Code;
	USHORT    GdtR0Data;

	USHORT    GdtR0Pcr;
	USHORT    GdtR3Code;
	USHORT    GdtR3Data;
	USHORT    GdtR3Teb;

	USHORT    GdtLdt;
	USHORT    GdtTss;
	USHORT    Gdt64R3CmCode;
	USHORT    Gdt64R3CmTeb;

	ULONG64   IopNumTriageDumpDataBlocks;
	ULONG64   IopTriageDumpDataBlocks;

	// Longhorn addition

	ULONG64   VfCrashDataBlock;
	ULONG64   MmBadPagesDetected;
	ULONG64   MmZeroedPageSingleBitErrorsDetected;

	// Windows 7 addition

	ULONG64   EtwpDebuggerData;
	USHORT    OffsetPrcbContext;

	// Windows 8 addition

	USHORT    OffsetPrcbMaxBreakpoints;
	USHORT    OffsetPrcbMaxWatchpoints;

	ULONG     OffsetKThreadStackLimit;
	ULONG     OffsetKThreadStackBase;
	ULONG     OffsetKThreadQueueListEntry;
	ULONG     OffsetEThreadIrpList;

	USHORT    OffsetPrcbIdleThread;
	USHORT    OffsetPrcbNormalDpcState;
	USHORT    OffsetPrcbDpcStack;
	USHORT    OffsetPrcbIsrStack;

	USHORT    SizeKDPC_STACK_FRAME;

	// Windows 8.1 Addition

	USHORT    OffsetKPriQueueThreadListHead;
	USHORT    OffsetKThreadWaitReason;

} KDDEBUGGER_DATA64, *PKDDEBUGGER_DATA64;

PKDDEBUGGER_DATA64 KdDebuggerDataBlock;

extern UCHAR *PsGetProcessImageFileName(IN PEPROCESS Process);

VOID InsertEntry(PEPROCESS pProcess)
{
	PLIST_ENTRY pList;
	pList = (PLIST_ENTRY)((ULONG64)pProcess + (ULONG)0x188);
	InsertTailList(PsActiveProcessHead, pList);
}

VOID RemoveEntry(PEPROCESS pProcess)
{
	PLIST_ENTRY pList;
	pList = (PLIST_ENTRY)((ULONG64)pProcess + (ULONG)0x188);
	RemoveEntryList(pList);
}

VOID CollectProcess(ULONG ProcessId, PEPROCESS pProcess)
{
	if (!IsAdded(wLastItem, pProcess))
		AddItem(&wLastItem, pProcess, ProcessId);
}

#define SUBVA64(base, offset)	(((ULONG64)(base) - (USHORT)(offset)))

BOOLEAN Initialize()
{
	KeSetSystemAffinityThreadEx(1);	// First core only

	KdDebuggerDataBlock = (PKDDEBUGGER_DATA64)(ULONG64)SUBVA64(HalDispatchTable, 0xbc0);

	PsActiveProcessHead = (PLIST_ENTRY)KdDebuggerDataBlock->PsActiveProcessHead;

	KeRevertToUserAffinityThread();

	DbgPrint("NtBuildLab: %s\r\n", KdDebuggerDataBlock->NtBuildLab);
	DbgPrint("PsActiveProcessHead : %I64x\n", PsActiveProcessHead);

	return TRUE;
}

#define PROCESS_CREATE_THREAD  (0x0002)
#define PROCESS_CREATE_PROCESS (0x0080)
#define PROCESS_TERMINATE      (0x0001)
#define PROCESS_VM_WRITE       (0x0020)
#define PROCESS_VM_READ        (0x0010)
#define PROCESS_VM_OPERATION   (0x0008)
#define PROCESS_SUSPEND_RESUME (0x0800)
#define PROCESS_QUERY_INFORMATION (0x0400)
#define PROCESS_DUP_HANDLE (0x0040)
#define NT_DEVICE_NAME      L"\\Device\\TestProject"
#define DOS_DEVICE_NAME     L"\\DosDevices\\TestProject"
#define MAXIMUM_FILENAME_LENGTH 256

PVOID _CallBacks_Handle = NULL;

typedef struct _OB_REG_CONTEXT
{
	__in USHORT Version;
	__in UNICODE_STRING Altitude;
	__in USHORT ulIndex;

	OB_OPERATION_REGISTRATION *OperationRegistration;

} REG_CONTEXT, *PREG_CONTEXT;

typedef PCHAR(*GET_PROCESS_IMAGE_NAME)(PEPROCESS Process);

LPSTR GetProcessNameFromPid(IN HANDLE Pid)
{
	PEPROCESS pProcess;

	if (PsLookupProcessByProcessId(Pid, &pProcess) == STATUS_INVALID_PARAMETER)
		return "NULL?!";

	return (LPSTR)PsGetProcessImageFileName(pProcess);
}

static const CHAR* PROTECTED_PROCESS_NAME = "ProtectedProcessTest.exe";

OB_PREOP_CALLBACK_STATUS ObjectPreCallback(IN PVOID RegistrationContext, IN POB_PRE_OPERATION_INFORMATION OperationInformation)
{
	HANDLE ProcessId = NULL;
	LPSTR ProcName = NULL;
	PEPROCESS pProcess = NULL;
	NTSTATUS ntStatus = STATUS_INVALID_PARAMETER;
	KIRQL OldIrql = 0;

	UNREFERENCED_PARAMETER(RegistrationContext);

	DbgPrintEx(DPFLTR_IHVDRIVER_ID, DPFLTR_ERROR_LEVEL, "ObjectPreCallback\n");

	ProcessId = PsGetProcessId((PEPROCESS)OperationInformation->Object);

	ProcName = GetProcessNameFromPid(ProcessId);

	DbgPrintEx(DPFLTR_IHVDRIVER_ID, DPFLTR_ERROR_LEVEL, "%s \n", ProcName);

	if (strstr(PROTECTED_PROCESS_NAME, ProcName) != NULL)
	{
		DbgPrintEx(DPFLTR_IHVDRIVER_ID, DPFLTR_ERROR_LEVEL, "ObjectPreCallback:: ProtectedProcessTest.exe \n");

		ntStatus = PsLookupProcessByProcessId(ProcessId, &pProcess);

		if (NT_SUCCESS(ntStatus))
		{
			if (!IsAdded(wLastItem, pProcess))	// if process is not already in tracking list...
			{
				AddItem(&wLastItem, pProcess, (ULONG)ProcessId);
				ObfDereferenceObject(pProcess);	// always after calling PsLookupProcessByProcessId 
				OldIrql = GlobalLock();
				RemoveEntry(pProcess);
				GlobalUnlock(OldIrql);
			}
			else
			{
				DbgPrintEx(DPFLTR_IHVDRIVER_ID, DPFLTR_ERROR_LEVEL, "denied second hide attempt\n");
			}
		}

		if (OperationInformation->Operation == OB_OPERATION_HANDLE_CREATE)
		{
			if ((OperationInformation->Parameters->CreateHandleInformation.OriginalDesiredAccess & PROCESS_TERMINATE) == PROCESS_TERMINATE)
			{
				OperationInformation->Parameters->CreateHandleInformation.DesiredAccess &= ~PROCESS_TERMINATE;
			}

			if ((OperationInformation->Parameters->CreateHandleInformation.OriginalDesiredAccess & PROCESS_VM_OPERATION) == PROCESS_VM_OPERATION)
			{
				OperationInformation->Parameters->CreateHandleInformation.DesiredAccess &= ~PROCESS_VM_OPERATION;
			}

			if ((OperationInformation->Parameters->CreateHandleInformation.OriginalDesiredAccess & ~PROCESS_VM_READ) == PROCESS_VM_READ)
			{
				OperationInformation->Parameters->CreateHandleInformation.DesiredAccess &= ~PROCESS_VM_READ;
			}

			if ((OperationInformation->Parameters->CreateHandleInformation.OriginalDesiredAccess & PROCESS_VM_WRITE) == PROCESS_VM_WRITE)
			{
				OperationInformation->Parameters->CreateHandleInformation.DesiredAccess &= ~PROCESS_VM_WRITE;
			}

			if ((OperationInformation->Parameters->CreateHandleInformation.OriginalDesiredAccess & PROCESS_QUERY_INFORMATION) == PROCESS_QUERY_INFORMATION)
			{
				OperationInformation->Parameters->CreateHandleInformation.DesiredAccess &= ~PROCESS_QUERY_INFORMATION;
			}

			if ((OperationInformation->Parameters->CreateHandleInformation.OriginalDesiredAccess & PROCESS_DUP_HANDLE) == PROCESS_DUP_HANDLE)
			{
				OperationInformation->Parameters->CreateHandleInformation.DesiredAccess &= ~PROCESS_DUP_HANDLE;
			}
		}
	}

	return OB_PREOP_SUCCESS;
}

VOID ObjectPostCallback(IN PVOID RegistrationContext, IN POB_POST_OPERATION_INFORMATION OperationInformation)
{
	UNREFERENCED_PARAMETER(RegistrationContext);
	UNREFERENCED_PARAMETER(OperationInformation);

	DbgPrintEx(DPFLTR_IHVDRIVER_ID, DPFLTR_ERROR_LEVEL, "ObjectPostCallback.\n");
}

NTSTATUS RegisterCallbackFunction()
{
	NTSTATUS ntStatus = STATUS_SUCCESS;

	UNICODE_STRING Altitude;

	USHORT filterVersion = ObGetFilterVersion();

	USHORT registrationCount = 1;

	OB_OPERATION_REGISTRATION RegisterOperation;

	OB_CALLBACK_REGISTRATION RegisterCallBack;

	REG_CONTEXT RegistrationContext;

	memset(&RegisterOperation, 0, sizeof(OB_OPERATION_REGISTRATION));

	memset(&RegisterCallBack, 0, sizeof(OB_CALLBACK_REGISTRATION));

	memset(&RegistrationContext, 0, sizeof(REG_CONTEXT));

	RegistrationContext.ulIndex = 1;

	RegistrationContext.Version = 120;

	if (filterVersion == OB_FLT_REGISTRATION_VERSION)
	{
		RegisterOperation.ObjectType = PsProcessType;

		RegisterOperation.Operations = OB_OPERATION_HANDLE_CREATE;

		RegisterOperation.PreOperation = ObjectPreCallback;

		RegisterOperation.PostOperation = ObjectPostCallback;

		RegisterCallBack.Version = OB_FLT_REGISTRATION_VERSION;

		RegisterCallBack.OperationRegistrationCount = registrationCount;

		RtlInitUnicodeString(&Altitude, L"XXXXXXX");

		RegisterCallBack.Altitude = Altitude;

		RegisterCallBack.RegistrationContext = &RegistrationContext;

		RegisterCallBack.OperationRegistration = &RegisterOperation;

		ntStatus = ObRegisterCallbacks(&RegisterCallBack, &_CallBacks_Handle);
	}

	return ntStatus;
}

NTSTATUS FreeProcFilter()
{
	if (NULL != _CallBacks_Handle)
	{
		ObUnRegisterCallbacks(_CallBacks_Handle);

		_CallBacks_Handle = NULL;
	}

	return STATUS_SUCCESS;
}

VOID DriverUnload(IN PDRIVER_OBJECT DriverObject)
{
	UNREFERENCED_PARAMETER(DriverObject);

	DbgPrintEx(DPFLTR_IHVDRIVER_ID, DPFLTR_ERROR_LEVEL, "DriverUnload: Driver unloaded.\n");

	FreeProcFilter();

	FreePointers(wLastItem);	// deallocate memory allocated before we unload our driver...
}

NTSTATUS DriverEntry(IN PDRIVER_OBJECT DriverObject, IN PUNICODE_STRING RegistryPath)
{
	NTSTATUS ntStatus;

	UNREFERENCED_PARAMETER(RegistryPath);

	DbgPrintEx(DPFLTR_IHVDRIVER_ID, DPFLTR_ERROR_LEVEL, "DriverEntry: Driver loaded. Use ed nt!Kd_IHVDRIVER_Mask f (or 7) to enable more traces\n");

	DriverObject->DriverUnload = DriverUnload;

	ntStatus = RegisterCallbackFunction();

	if (!NT_SUCCESS(ntStatus))
		return ntStatus;

	if (!Initialize())
		return STATUS_NOT_SUPPORTED;

	return STATUS_SUCCESS;
}